import { escape } from 'lodash';
import { join } from 'path';
import * as RelateUrl from 'relateurl';
import { combineLatest, Observable, Subject } from 'rxjs';
import { map, mergeMap } from 'rxjs/operators';
import { parse as urlParse } from 'url';

import { firstOf } from '../utils/file.utils';
import {
    CMD_ADD,
    OPTION_DATA,
    OPTION_ESI,
    OPTION_NAME,
    OPTION_RELATIVE,
    OPTION_ROOT,
    OPTION_SOURCE,
    OPTION_URL,
    OPTION_VERBOSE,
} from './../constants';
import { Options, optNormalizeDir } from './../options';
import { configIndexDirFromContextRoot, configTrimSlashes } from './../utils/config';
import { readFile as rxReadFile, writeFile as rxWriteFile } from './../utils/rx.walk';
import { ensureTrailingSlash } from './../utils/url.utils';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import { KEY_MANIFEST_ID } from './create.webappmanifest.command';
import {
    HELP_API_URL,
    HELP_APPLICATION_NAME,
    HELP_CONTEXT_ROOT,
    HELP_DATA_DIRECTORY,
    HELP_ESI,
    HELP_INDEX_SOURCE,
    HELP_RELATIVE,
    HELP_VERBOSE,
} from './help';

/* Copyright IBM Corp. 2017 */
const HEAD_REGEXP = /<\/head>/g;
const FAVICON_REGEXP = /<link\s+[^>]*rel\s*=\s*\"icon\"[^>]*>[\s|\r|\n]*/g;
const MANIFEST_REGEXP = /<link\s+[^>]*rel\s*=\s*\"manifest\"[^>]*>[\s|\r|\n]*/g;

function _createUrl(aBaseUrl: RelateUrl, aDeliveryBase: string, aRelUrl: string, bRelative: boolean): string {
    // parse
    const url = `${aDeliveryBase}${aRelUrl}`;
    const parsed = urlParse(url);
    return (bRelative ? aBaseUrl.relate(url) : `//${parsed.host}${parsed.path}`) || url;
}

/**
 * Creates the links to the manifest file and the favicons
 *
 * @param aDeliveryBase base URL
 * @param aManifestType manifest type
 * @param bRelative check if the links should be relative or absolute
 * @param aLogger   the logger
 */
function _createLinks(aDeliveryBase: string,
    aContextRoot: string,
    aManifestType: string,
    bRelative: boolean,
    bEsi: boolean,
    aLogger: Subject<string>): string {
    // log this
    aLogger.next(`Delivery base [${aDeliveryBase}].`);
    // the base URL
    const absRootUrl = ensureTrailingSlash(`${aDeliveryBase}${configTrimSlashes(aContextRoot)}`);
    // base URL
    const baseUrl = new RelateUrl(absRootUrl);
    // URL to the manifest
    const manifestUrl = _createUrl(baseUrl, aDeliveryBase, `templates/manifest/${aManifestType}_manifest.json`, bRelative);
    // URL to the favicons
    const faviconsUrl = _createUrl(baseUrl, aDeliveryBase, `templates/favicons/${aManifestType}_favicons.html`, bRelative);
    // produce the markup
    const links: string[] = [];
    links.push(`<link rel="manifest" href="${escape(manifestUrl)}">`);
    // check if we would create ESI or not
    if (bEsi) {
        links.push(`<esi:include src="${escape(faviconsUrl)}"></esi:include>`);
    } else {
        links.push(`<link rel="esi:include" href="${escape(faviconsUrl)}">`);
    }
    // ok
    return links.join('\n');
}

function _insertMarkup(aFavicons: string, aMarkup: string): string {
    // delete an existing icon
    let html = aMarkup.replace(FAVICON_REGEXP, '').replace(MANIFEST_REGEXP, '');
    // replace the head regular expression
    html = html.replace(HEAD_REGEXP, `${aFavicons}\n</head>`);
    // ok
    return html;
}

function _rewriteIndex(aMarkup: string, aSrcDir: string): Observable<string> {
    // read the index file
    const fileName = join(aSrcDir, 'index.html');
    return rxReadFile(fileName).pipe(
        map(content => _insertMarkup(aMarkup, content)),
        mergeMap(newMarkup => rxWriteFile(fileName, newMarkup))
    );
}

function _addWebAppManifest(aDeliveryBase: string,
    aContextRoot: string,
    aManifestType: string,
    bRelative: boolean,
    bEsi: boolean,
    aSrcDir: string,
    aLogger: Subject<string>): Observable<string> {
    // construct the markup
    const markup = _createLinks(aDeliveryBase, aContextRoot, aManifestType, bRelative, bEsi, aLogger);
    // rewrite
    return _rewriteIndex(markup, aSrcDir);
}

export class AddWebAppManifestCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'webappmanifest';

    static readonly COMMAND = CMD_ADD;

    constructor(options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(AddWebAppManifestCommand.COMMAND, AddWebAppManifestCommand.TYPE,
            `[--${OPTION_SOURCE} <SOURCE>]`,
            `[--${OPTION_DATA} <DATA>]`,
            `[--${OPTION_URL} <URL>]`,
            `[--${OPTION_NAME} <NAME>]`,
            `[--${OPTION_ROOT} <ROOT>]`,
            `[--${OPTION_RELATIVE}]`,
            `[--${OPTION_ESI}]`,
            `[--${OPTION_VERBOSE}]`);
        // help on the options
        console.log(`--${OPTION_SOURCE} <SOURCE>`, ':', HELP_INDEX_SOURCE);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_URL} <URL>`, ':', HELP_API_URL);
        console.log(`--${OPTION_NAME} <NAME>`, ':', HELP_APPLICATION_NAME);
        console.log(`--${OPTION_ROOT}`, ':', HELP_CONTEXT_ROOT);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
        console.log(`--${OPTION_RELATIVE}`, ':', HELP_RELATIVE);
        console.log(`--${OPTION_ESI}`, ':', HELP_ESI);
    }

    process(): Observable<string> {
        // logger
        const logger = this.getLogger();
        // data directory
        const opts = this.getOptions();
        // the directories
        const dataDirOpt = this.getDataDir();
        const appName = opts.name || this.getAppName();
        const contextRoot = `/${configTrimSlashes(opts.root || this.getContextRoot(appName))}`;
        // the directories
        const srcDirOpt = optNormalizeDir(firstOf(opts.src) || configIndexDirFromContextRoot(contextRoot, dataDirOpt));
        const rxSrcDir = this.rxMkdirp(srcDirOpt);
        // the url
        const rxDeliveryBase = this.getDeliveryBase();
        // type id
        const manifestId = opts.id || this.getConfigValue(KEY_MANIFEST_ID) || '';
        // relative or absolute
        const bRelative = !!opts.relative;
        // ESI
        const bEsi = !!opts.esi;
        // proceed
        logger.next(`Source dir [${srcDirOpt}], manifestid [${manifestId}], relative [${bRelative}], esi [${bEsi}].`);
        // proceed
        return combineLatest(rxSrcDir, rxDeliveryBase).pipe(
            mergeMap(([aSrcDir, aDeliveryBase]) =>
                _addWebAppManifest(aDeliveryBase, contextRoot, manifestId, bRelative, bEsi, aSrcDir, logger)
            ));
    }

}
