/* Copyright IBM Corp. 2017 */
import * as Assert from 'assert';
import { isString } from 'lodash';
import * as RelateUrl from 'relateurl';
import { combineLatest, empty, Observable } from 'rxjs';
import { catchError, filter, map, shareReplay, switchMap } from 'rxjs/operators';

import { ensureTrailingSlash, wchDeliveryUrlFromApiUrl } from '../utils/wchtools';
import { CMD_APP, OPTION_DATA, OPTION_NAME, OPTION_PASSWORD, OPTION_URL, OPTION_USER, OPTION_VERBOSE } from './../constants';
import { Options } from './../options';
import { configGetBaseUrl, configTrimSlashes } from './../utils/config';
import { AbstractWchToolsCommand } from './abstract.wchtools.command';
import {
    HELP_API_URL,
    HELP_APPLICATION_NAME,
    HELP_DATA_DIRECTORY,
    HELP_PASSWORD,
    HELP_USERNAME,
    HELP_VERBOSE,
} from './help';

export function assertHasContext(aValue: string): string {
    Assert.ok(isString(aValue) && (aValue.length > 0), 'Application has to use a context root.');
    return aValue;
}

export class ApplicationPurgeCommand extends AbstractWchToolsCommand {

    static readonly TYPE = 'purge';

    static readonly COMMAND = CMD_APP;

    constructor(aOptions: Options) {
        super(aOptions);
    }

    printHelp(): void {
        // log the command line
        console.log(ApplicationPurgeCommand.COMMAND, ApplicationPurgeCommand.TYPE,
            `--${OPTION_DATA} <DATA>`,
            `[--${OPTION_NAME} <NAME>]`,
            `[--${OPTION_URL} <URL>]`,
            `[--${OPTION_USER} <USER>]`,
            `[--${OPTION_PASSWORD} <PASSWORD>]`,
            `[--${OPTION_VERBOSE}]`);
        // log the parameters
        console.log(`--${OPTION_NAME} <NAME>`, ':', HELP_APPLICATION_NAME);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_URL} <URL>`, ':', HELP_API_URL);
        console.log(`--${OPTION_USER} <USER>`, ':', HELP_USERNAME);
        console.log(`--${OPTION_PASSWORD} <PASSWORD>`, ':', HELP_PASSWORD);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // the options
        const opts = this.getOptions();
        // options
        const appName = opts.name || this.getAppName();
        const contextRoot = `/${configTrimSlashes(opts.root || this.getContextRoot(appName))}`;
        // api base
        const rxApiBaseUrl = this.getApiBase().pipe(
            shareReplay()
        );
        // the delivery URL
        const rxAbsWchDeliveryUrl = rxApiBaseUrl.pipe(
            map(aApiBaseUrl => ensureTrailingSlash(wchDeliveryUrlFromApiUrl(aApiBaseUrl))),
            shareReplay()
        );
        // path of the index file
        const rxAbsRootUrl = rxAbsWchDeliveryUrl.pipe(
            map(absWchDeliveryUrl => ensureTrailingSlash(`${absWchDeliveryUrl}${configTrimSlashes(contextRoot)}`)),
            shareReplay()
        );
        // get the base URL
        const rxAbsAppUrl = rxApiBaseUrl.pipe(
            map(aApiBaseUrl => ensureTrailingSlash(configGetBaseUrl(appName, aApiBaseUrl))),
            shareReplay()
        );
        // relative delivery
        const rxDeliveryUrl = combineLatest(rxAbsRootUrl, rxAbsAppUrl).pipe(
            map(([absRootUrl, absAppUrl]) => new RelateUrl(absRootUrl).relate(absAppUrl)),
            shareReplay()
        );
        // construct the url to delete
        const rxPathToDelete = rxDeliveryUrl.pipe(
            map(configTrimSlashes),
            map(assertHasContext),
            filter(Boolean),
            map(url => '/' + url),
            shareReplay()
        );
        // construct the command line arguments
        const rxArgs = rxPathToDelete.pipe(
            map(path => ['delete', '-w', '--recursive', '--path', path, '--quiet'])
        );

        // execute
        return rxArgs.pipe(
            switchMap(args => this.executeWchTools(args)),
            catchError(err => empty())
        );
    }
}
