/* Copyright IBM Corp. 2017 */
import { isString } from 'lodash';
import { getType } from 'mime';
import { Observable } from 'rxjs';
import { filter, map, mergeMap } from 'rxjs/operators';
import { isArray } from 'util';

import { canonicalizeJSON } from '../utils/json';
import { addRevisionId } from '../utils/revision.utils';
import { Options } from './../options';
import { JsonFile, readJSON, writeJSON } from './../utils/rx.json';
import { walkFiles } from './../utils/rx.walk';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';

const _isArray = Array.isArray;
const _keys = Object.keys;

type Handler = (aDoc: any, aKey: string, aParent: any) => void;

function _delete(aDoc: any, aKey: string, aParent: any) {
    delete aParent[aKey];
}

interface HandlerMap {
    [key: string]: Handler;
}

function _cleanupThumbnail(aThumb: any, aKey: string, aParent: any) {
    // ID is sufficient
    if (isString(aThumb.id)) {
        delete aThumb.path;
    }
    delete aThumb.url;
}

function _cleanupElement(aElement: any, aKey: string, aParent: any) {

}

function _cleanupElements(aElement: any, aKey: string, aParent: any) {

}

function _deleteName(aType: any, aKey: string | number, aParent: any) {
    if (aType && isString(aType.id)) {
        delete aType.name;
    }
}

function _cleanupMapping(aData: any, aKey: number, aParent: any) {
    _deleteName(aData.defaultLayout, 'defaultLayout', aData);
    const layouts = aData.layouts;
    if (isArray(layouts)) {
        layouts.forEach(_deleteName);
    }
}

function _cleanupMappings(aData: any, aKey: string, aParent: any) {
    if (isArray(aData)) {
        aData.forEach(_cleanupMapping);
    }
}

const _GENERIC_HANDLERS: HandlerMap = {
    'rev': _delete,
    'publishing': _delete,
    'creatorId': _delete,
    'lastModifierId': _delete,
    'created': _delete,
    'lastModified': _delete,
    'creator': _delete,
    'lastModifier': _delete,
    'links': _delete,
    'cognitive': _delete
};

const _TYPE_HANDLERS: HandlerMap = Object.assign({
    'thumbnail': _cleanupThumbnail,
    'elements': _cleanupElements
}, _GENERIC_HANDLERS);

const _LAYOUT_HANDLERS: HandlerMap = Object.assign({
    'thumbnail': _cleanupThumbnail
}, _GENERIC_HANDLERS);

const _LAYOUT_MAPPING_HANDLERS: HandlerMap = Object.assign({
    'type': _deleteName,
    'mappings': _cleanupMappings
}, _GENERIC_HANDLERS);

const _SITE_HANDLERS: HandlerMap = Object.assign({
}, _GENERIC_HANDLERS);

const _PAGE_HANDLERS: HandlerMap = Object.assign({
    // 'contentTypeId': _delete
}, _GENERIC_HANDLERS);

const _CONTENT_HANDLERS: HandlerMap = Object.assign({
    'thumbnail': _cleanupThumbnail,
    // 'contentTypeId': _delete
}, _GENERIC_HANDLERS);

const _DELETE_TYPE = ['rev', 'creatorId', 'lastModifierId', 'created', 'lastModified', 'creator', 'lastModifier', 'links', 'cognitive'];

function _cleanupAsset(aAsset: any, aHandlerMap: HandlerMap) {
    _keys(aAsset).filter(k => !!aHandlerMap[k]).forEach(k => aHandlerMap[k](aAsset[k], k, aAsset));
}

function _cleanupType(aType: any) {
    _cleanupAsset(aType, _TYPE_HANDLERS);
}

function _cleanupLayout(aType: any) {
    _cleanupAsset(aType, _LAYOUT_HANDLERS);
}

function _cleanupLayoutMappings(aType: any) {
    _cleanupAsset(aType, _LAYOUT_MAPPING_HANDLERS);
}

function _cleanupSite(aType: any) {
    _cleanupAsset(aType, _SITE_HANDLERS);
}

function _cleanupPage(aType: any) {
    _cleanupAsset(aType, _PAGE_HANDLERS);
}

function _cleanupContent(aType: any) {
    if (isString(aType.typeId)) {
        delete aType.type;
    }
    _cleanupAsset(aType, _CONTENT_HANDLERS);
}

const _DELETE = ['rev', 'creatorId', 'lastModifierId', 'created', 'lastModified', 'creator', 'lastModifier', 'links', 'cognitive'];

/**
 * Tests for a plain object
 *
 * @param aObject   the object to test
 */
function _isPlainObject(aObject: any): boolean {
    return aObject && !aObject.prototype && (typeof aObject === 'object');
}

function _cleanUp(aData: any): any {
    // delete
    if (_isPlainObject(aData)) {
        switch (aData.classification) {
            case 'content-type':
                _cleanupType(aData);
                break;
            case 'layout':
                _cleanupLayout(aData);
                break;
            case 'layout-mapping':
                _cleanupLayoutMappings(aData);
                break;
            case 'site':
                _cleanupSite(aData);
                break;
            case 'page':
                _cleanupPage(aData);
                break;
            case 'content':
                _cleanupContent(aData);
                break;
        }
    }
    // canonicalize
    return canonicalizeJSON(addRevisionId(aData));
}

function _cleanupJSON(aFile: JsonFile): JsonFile {
    aFile.data = _cleanUp(aFile.data);
    return aFile;
}

export class CleanupAssetsCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'cleanup';
    static readonly COMMAND = 'assets';

    constructor(options: Options) {
        super(options);
    }

    process(): Observable<string> {
        // data dir
        const rxDataDir = this.rxMkdirp(this.getDataDir());

        // process
        return rxDataDir.pipe(
            mergeMap(walkFiles),
            filter(f => getType(f.absPath) === 'application/json'),
            mergeMap(readJSON),
            map(_cleanupJSON),
            mergeMap(writeJSON),
            map(file => file.absPath)
        );
    }
}

