/* Copyright IBM Corp. 2018 */
import { join, normalize, parse } from 'path';
import { cwd } from 'process';
import { concat, merge, MonoTypeOperatorFunction, Observable, of, Subject } from 'rxjs';
import { distinct, map, mergeMap, mergeMapTo, shareReplay } from 'rxjs/operators';
import { parse as parseUrl } from 'url';

import { CMD_CREATE, OPTION_ARCHIVE } from '../constants';
import { assertIsUrl, assertParameter } from '../utils/assert';
import { firstOf } from '../utils/file.utils';
import { canonicalizeJSON } from '../utils/json';
import { rxDeleteDir } from '../utils/rx.delete';
import { rxReadTextFile, rxWriteTextFile } from '../utils/rx.file';
import { readFile, writeFile } from '../utils/rx.walk';
import { rxUnzipFromUrl } from '../utils/rx.zip';
import { Options, optNormalizeDir } from './../options';
import { TENANT_BASED_URL_REGEXP } from './../utils/tenant';
import { kebabCase } from './../utils/transliteration';
import { WchToolsOptions } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import {
    createProjectCopyResources,
    createProjectCreateWchtoolsOptions,
    createProjectUpdateTypings,
} from './create.project.command';

// relative path to the data directory
const REL_DATA_DIR = './src/wchLayouts';
const REL_SOURCE_DIR = './src';

const DOMAIN_NAME_REGEXP = /(?:\/\/\s*)(.+DOMAIN_NAME\s+=\s+['"])[^'"]+(["'];)/;
const HUB_ID_REGEXP = /(?:\/\/\s*)(.+CONTENT_HUB_ID\s+=\s+['"])[^'"]+(["'];)/;

function _rewriteFiles(aName: string, aWchToolsOptions: WchToolsOptions, aLogger: Subject<string>): MonoTypeOperatorFunction<string> {

    function _updateConstants(aData: string): string {
        // parse the url
        const parsedUrl = parseUrl(aWchToolsOptions.baseUrl);
        // replace the domain name
        let text = aData;
        if (parsedUrl.host) {
            text = text.replace(DOMAIN_NAME_REGEXP, (total, prefix, suffix) => `${prefix}${parsedUrl.host}${suffix}`);
        }
        // tenant ID
        if (parsedUrl.pathname) {
            const result = TENANT_BASED_URL_REGEXP.exec(parsedUrl.pathname);
            if (result && (result.length > 1)) {
                // extract the tenant
                const tenantID = result[1];
                text = text.replace(HUB_ID_REGEXP, (total, prefix, suffix) => `${prefix}${tenantID}${suffix}`);
            }
        }
        // ok
        return text;
    }

    function _rewriteConstants(aFile: string): Observable<string> {
        return rxReadTextFile(aFile).pipe(
            map(data => _updateConstants(data)),
            mergeMap(data => rxWriteTextFile(aFile, data))
        );
    }

    function _rewriteTypings(aFile: string): Observable<string> {
        return createProjectUpdateTypings(aFile, aLogger);
    }

    function _updatePackage(aPackage: any): any {
        // add a config section
        const cfg = aPackage.config || (aPackage.config = {});
        // update
        cfg.root = '/';
        cfg.data = REL_DATA_DIR;
        cfg.src = REL_SOURCE_DIR;
        cfg.name = aName;
        // update the application name
        aPackage.name = aName;

        return aPackage;
    }

    function _rewritePackageJson(aFile: string): Observable<string> {
        return readFile(aFile).pipe(
            map(data => JSON.parse(data)),
            map(pkg => _updatePackage(pkg)),
            map(canonicalizeJSON),
            map(json => JSON.stringify(json, null, 2)),
            mergeMap(pkg => writeFile(aFile, pkg))
        );
    }

    function _rewriteFile(aFile: string): Observable<string> {
        // parse
        const parsed = parse(aFile);
        // test some special cases
        switch (parsed.base) {
            case 'package.json':
                return _rewritePackageJson(aFile);
            case 'typings.d.ts':
                return _rewriteTypings(aFile);
            case 'Constants.ts':
                return _rewriteConstants(aFile);
            default:
                return of(aFile);
        }
    }

    // the rewrite operator
    return (aSrc: Observable<string>): Observable<string> => {
        // returns the rewrite operators
        return aSrc.pipe(mergeMap(_rewriteFile));
    };
}

function _extractFromUrl(aSrcUrl: string, aDstDir: string, aLogger: Subject<string>): Observable<string> {
    // log this
    aLogger.next(`Extracting [${aSrcUrl}] to [${aDstDir}].`);
    // just dispatch to the magic extractor, skip the first target dir
    return rxUnzipFromUrl(aSrcUrl, aDstDir, 1);
}

export class CreateFromArchiveCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'fromarchive';

    static readonly COMMAND = CMD_CREATE;

    constructor(private options: Options) {
        super(options);
    }

    process(): Observable<string> {
        // options
        const opts = this.options;
        // archive
        const archive = opts.archive;
        assertParameter(archive, OPTION_ARCHIVE);
        assertIsUrl(archive);
        // switch to the target directory
        const srcDir = optNormalizeDir(firstOf(opts.src) || '.');
        const rxSrcDir = this.rxMkdirp(srcDir);
        // capture the current working directory
        const currentDir = cwd();
        const name = kebabCase(opts.name || this.getAppName());
        const dstDir = join(srcDir, name);
        // pull content?
        const bSkipPull = opts.skipPull || false;
        // check if we would override
        const rxClean = (this.options.override && name)
            ? rxSrcDir.pipe(mergeMap(aSrcDir => rxDeleteDir(dstDir).pipe(mergeMapTo(rxSrcDir))))
            : rxSrcDir;

        // data directory
        const dataDir = normalize(join(dstDir, REL_DATA_DIR));

        // logger
        const logger = this.getLogger();

        // base URL
        const rxToolsOpts = this.getWchToolsOptions().pipe(shareReplay(1));

        // check if we create an inline edit experience
        const bEditable = opts.editable || false;

        // write the wchtools config
        const rxCreateWchtoolsOptions = rxToolsOpts.pipe(mergeMap(cfg => createProjectCreateWchtoolsOptions(dataDir, cfg, logger)));

        // copy some standard resources
        const rxCopyResources = rxToolsOpts.pipe(mergeMap(cfg => createProjectCopyResources(srcDir, name, cfg, logger)));

        const rxRewriteOperator = rxToolsOpts.pipe(map(cfg => _rewriteFiles(name, cfg, logger)));

        // extract the files
        const rxExtract = rxClean.pipe(mergeMap(() => _extractFromUrl(archive || '', dstDir, logger)));

        const rxExtractAndRewrite = rxRewriteOperator.pipe(mergeMap(op => rxExtract.pipe(op)));

        const rxAfterExtract = merge(rxCreateWchtoolsOptions, rxCopyResources);

        // total
        const rxTotal = concat(rxExtractAndRewrite, rxAfterExtract).pipe(distinct());

        // construct the target
        return rxTotal;
    }
}

