/* Copyright IBM Corp. 2017 */
import { camelCase, toLower, upperFirst } from 'lodash';
import { getType } from 'mime';
import { join } from 'path';
import { concat, empty, merge, Observable, of, Subject } from 'rxjs';
import { filter, map, mergeMap, shareReplay } from 'rxjs/operators';
import { v4 } from 'uuid';

import {
    CMD_CREATE,
    OPTION_DATA,
    OPTION_FLAT,
    OPTION_OVERRIDE,
    OPTION_SCSS,
    OPTION_SOURCE,
    OPTION_TYPE,
    OPTION_VERBOSE,
} from './../constants';
import { Options } from './../options';
import { canonicalizeJSON } from './../utils/json';
import { addRevisionId } from './../utils/revision.utils';
import { readFile as rxReadFile, rxMkdirp, writeFile as rxWriteFile } from './../utils/rx.walk';
import { kebabCase } from './../utils/transliteration';
import { WchTools } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import { CreateAssetCommand } from './create.asset.command';
import { CreateLayoutCommand } from './create.layout.command';
import {
    HELP_BUILD_SOURCE,
    HELP_DATA_DIRECTORY,
    HELP_FLAT,
    HELP_LAYOUT_TYPE,
    HELP_OVERRIDE,
    HELP_SCSS,
    HELP_VERBOSE,
} from './help';

function _fixPageType(aType: any, aName: string, aThumbId: string, aTools: WchTools): any {
    // construct the id
    const id = _getType(aName, aTools) || v4();
    // add ID
    aType.id = id;
    aType.name = aName;
    // type name
    aType.thumbnail = {
        id: aThumbId
    };
    // inject the revision field
    return canonicalizeJSON(addRevisionId(aType));
}

function _createPageTypeDoc(aName: string, aThumbId: string, aTools: WchTools): Observable<any> {
    // load the template
    return rxReadFile(join(__dirname, '..', '..', '..', 'assets', 'page.type.json')).pipe(
        map(data => JSON.parse(data)),
        map(template => _fixPageType(template, aName, aThumbId, aTools))
    );
}

function _getType(aName: string, aTools: WchTools): string | undefined {
    // iterate
    // tslint:disable-next-line:forin
    for (const id in aTools.types) {
        // content type
        const doc = aTools.types[id];
        if (doc && doc.name === aName) {
            return id;
        }
    }
}

function _hasType(aName: string, aTools: WchTools): boolean {
    return !!_getType(aName, aTools);
}

function _writeThumbnail(aName: string, aDataDir: string, aOpts: Options, aLogger: Subject<string>): Observable<string> {
    // id
    const id = kebabCase(aName);
    // target image
    const srcThumb = join(__dirname, '..', '..', '..', 'assets', 'page.thumbnail.jpg');
    // construct an asset for the thumbnail
    return of({
        command: CreateAssetCommand.COMMAND,
        commandType: CreateAssetCommand.TYPE,
        data: aDataDir,
        src: srcThumb,
        name: `${id}.jpg`,
        verbose: aOpts.verbose,
        allArgs: aOpts.allArgs
    }).pipe(
        map(opts => new CreateAssetCommand(opts)),
        mergeMap(cmd => cmd.process())
    );
}

function _isJSON(aName: string): boolean {
    return !!(aName && (getType(aName) === 'application/json'));
}

function _getThumbId(aFiles: Observable<string>): Observable<string> {
    // filter down to JSON files
    return aFiles.pipe(
        filter(_isJSON),
        filter(name => (name.indexOf('_amd') >= 0)),
        mergeMap(file => rxReadFile(file)),
        map(data => JSON.parse(data)),
        map(data => data.id)
    );
}

/**
 * Produces a type document in the data directory
 *
 * @param aName     the name of the type
 * @param aDataDir  the data directory
 */
function _writePageType(aName: string, aDataDir: string, aTools: WchTools,
    aOverride: boolean,
    aOptions: Options,
    aLogger: Subject<string>): Observable<string> {
    // check if we know the type
    if (!aOverride && _hasType(aName, aTools)) {
        return empty();
    }
    // construct the target directory
    const dstDir = join(aDataDir, 'types');
    // thumb
    const thumb = _writeThumbnail(aName, aDataDir, aOptions, aLogger).pipe(shareReplay(1));
    // find the ID of the generated thumb
    const thumbId = _getThumbId(thumb);
    // produce
    const json = rxMkdirp(dstDir).pipe(
        mergeMap(dir =>
            thumbId.pipe(mergeMap(thId => _createPageTypeDoc(aName, thId, aTools).pipe(
                mergeMap(doc => rxWriteFile(join(dir, `${doc.id}_tmd.json`), JSON.stringify(doc, undefined, 2))
                )))))
    );
    // merge
    return merge(thumb, json);
}

function _createLayout(aName: string, aOpt: Options, aLogger: Subject<string>): Observable<string> {
    // log this
    aLogger.next('Dispatching to CreateLayoutCommand');
    // operate
    return of({
        data: aOpt.data,
        src: aOpt.src,
        type: aName,
        verbose: aOpt.verbose,
        flat: aOpt.flat,
        scss: aOpt.scss,
        module: aOpt.module,
        skipRegistration: aOpt.skipRegistration,
        command: CreateLayoutCommand.COMMAND,
        commandType: CreateLayoutCommand.TYPE,
        allArgs: aOpt.allArgs
    }).pipe(
        map(opts => new CreateLayoutCommand(opts)),
        mergeMap(cmd => cmd.process())
    );
}

const SUFFIX = 'type';

function _fixTypeName(aName: string): string {
    // check for the lower case version
    let name = toLower(aName);
    if (name.endsWith(SUFFIX)) {
        name = aName.substring(0, aName.length - SUFFIX.length);
    }
    // append the suffix
    name = `${name} ${SUFFIX}`;
    // camel case the beast
    return upperFirst(camelCase(name));
}

/**
 * create pagetype --data <DATA> --src <SRC> --type <TYPE> [--verbose] [--flat] [--scss] [--override]
 */
export class CreatePageTypeCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'pagetype';

    static readonly COMMAND = CMD_CREATE;

    constructor(private options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(CreatePageTypeCommand.COMMAND, CreatePageTypeCommand.TYPE,
            `--${OPTION_SOURCE} <SOURCE>`,
            `--${OPTION_DATA} <DATA>`,
            `--${OPTION_TYPE} <TYPE>`,
            `[--${OPTION_FLAT}]`,
            `[--${OPTION_SCSS}]`,
            `[--${OPTION_OVERRIDE}]`,
            `[--${OPTION_VERBOSE}]`);
        // log the parameters
        console.log(`--${OPTION_SOURCE} <SOURCE>`, ':', HELP_BUILD_SOURCE);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_TYPE} <TYPE>`, ':', HELP_LAYOUT_TYPE);
        console.log(`--${OPTION_FLAT}`, ':', HELP_FLAT);
        console.log(`--${OPTION_SCSS}`, ':', HELP_SCSS);
        console.log(`--${OPTION_OVERRIDE}`, ':', HELP_OVERRIDE);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // access to the data folders
        const opts = this.options;
        // the directories
        const dataDirOpt = this.getDataDir();
        const rxDataDir = this.rxMkdirp(dataDirOpt);
        // logger
        const logger = this.getLogger();
        // override flag
        const override = !!opts.override;
        // the type
        const type = _fixTypeName(opts.type || 'PageType');
        // tools
        const rxWchTools = this.getWchTools();
        // log this
        logger.next(`Data dir [${dataDirOpt}], type [${type}].`);
        // done
        return rxDataDir.pipe(
            mergeMap(aDataDir => rxWchTools.pipe(mergeMap(aWchTools => concat(
                _writePageType(type, aDataDir, aWchTools, override, opts, logger),
                _createLayout(type, opts, logger)
            )))));
    }
}
