/* Copyright IBM Corp. 2017 */
import { camelCase, find } from 'lodash';
import { join } from 'path';
import { combineLatest, concat, empty, merge, Observable, of, Subject } from 'rxjs';
import { distinct, map, mergeMap, shareReplay, tap } from 'rxjs/operators';
import { v4 } from 'uuid';

import { firstOf } from '../utils/file.utils';
import { CMD_CREATE } from './../constants';
import { Options, optNormalizeDir } from './../options';
import { canonicalizeJSON } from './../utils/json';
import { addRevisionId } from './../utils/revision.utils';
import { rxCopyDir } from './../utils/rx.copy';
import { readFile as rxReadFile, rxMkdirp, writeFile as rxWriteFile } from './../utils/rx.walk';
import { WchTools } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';

export const UUID_MANIFEST_TYPE = '3900cd47-0f9a-45a9-892f-3f9a3e42c751';
export const UUID_MANIFEST_LAYOUT = '72b59cc2-3664-4a55-a8df-3604b503ed47';
export const UUID_FAVICONS_LAYOUT = '3fb75fda-4ff2-42a0-be53-69837df22f43';

export const KEY_MANIFEST_ID = 'manifestId';

const CONTENT_REGEXP = /content.*_cmd\.json/g;

interface State {
    contentItemId?: string;
}

function _isContent(aFileName: string): boolean {
    return CONTENT_REGEXP.test(aFileName);
}

function _getLayoutMappingForType(aTypeId: string, aTools: WchTools): any {
    const foundId = Object.keys(aTools.layoutMappings).find(id => aTools.layoutMappings[id].type.id === aTypeId);
    return foundId ? aTools.layoutMappings[foundId] : undefined;
}

function _findManifestLayout(aTools: WchTools, aLogger: Subject<string>): any {
    // locate
    return find(aTools.layouts, layout => layout.id === UUID_MANIFEST_LAYOUT);
}

function _findFaviconsLayout(aTools: WchTools, aLogger: Subject<string>): any {
    // locate
    return find(aTools.layouts, layout => layout.id === UUID_FAVICONS_LAYOUT);
}

function _createLayoutMapping(aTools: WchTools, aLogger: Subject<string>): Observable<string> {
    // new ID
    const id = v4();
    // load the template
    const srcFile = join(__dirname, '..', '..', '..', 'assets', 'manifest.layout.mapping.json');
    const dstFolder = join(aTools.root, 'layout-mappings');
    const dstFile = join(dstFolder, `${id}.json`);
    // create the folder
    const rxDstFolder = rxMkdirp(dstFolder);
    // log this
    aLogger.next(`Creating layout mapping with ID ${id} ...`);
    // load the file
    const rxMapping = rxReadFile(srcFile).pipe(
        map(data => JSON.parse(data)),
        map(mapping => {
            // update
            mapping.id = id;
            mapping.name = id;
            mapping.type = {
                id: UUID_MANIFEST_TYPE
            };
            // ok
            const newMapping = canonicalizeJSON(addRevisionId(mapping));
            aTools.layoutMappings[id] = newMapping;
            aTools.idToFile[id] = dstFile;
            // ok
            return newMapping;
        })
    );
    // combine
    return combineLatest(rxMapping, rxDstFolder).pipe(
        mergeMap(([mapping]) => rxWriteFile(dstFile, JSON.stringify(mapping, undefined, 2))),
        shareReplay()
    );
}

function _addLayoutToMappings(aLayoutID: string, aMappings: any, aTools: WchTools) {
    // add to existing layouts
    const layouts = aMappings.layouts || (aMappings.layouts = []);
    const exLayout = layouts.find((m: any) => m.id === aLayoutID);
    // add the layout if it does not exist
    if (!exLayout) {
        layouts.push({
            id: aLayoutID
        });
    }
    // add the default
    const defLayout = aMappings.defaultLayout;
    if (!defLayout) {
        aMappings.defaultLayout = {
            id: aLayoutID
        };
    }
}

function _addLayoutToMapping(aLayoutID: string,
    aMapping: any,
    aTools: WchTools,
    aLogger: Subject<string>): Observable<string> {
    // log this
    aLogger.next(`_addLayoutToMapping [${aLayoutID}] ...`);
    // find the layout
    const mappings = aMapping.mappings || (aMapping.mappings = []);
    // iterate
    mappings.forEach((m: any) => _addLayoutToMappings(aLayoutID, m, aTools));
    // handle the empty case
    if (mappings.length === 0) {
        const m = {};
        _addLayoutToMappings(aLayoutID, m, aTools);
        mappings.push(m);
    }
    // write the beast
    const fileName = aTools.idToFile[aMapping.id];
    // update the mappings
    aTools.layoutMappings[aMapping.id] = aMapping;
    // write
    return rxWriteFile(fileName, JSON.stringify(aMapping, undefined, 2));
}

function _fixSlashes(aPath: string): string {
    // replace the slashes
    let p = aPath.replace(/\\/g, '/');
    if (!p.startsWith('/')) {
        p = '/' + p;
    }
    // ok
    return p;
}

function _writeLayoutAndMapping(aRoot: string, aTypeId: string,
    aLayoutId: string, aTools: WchTools, aLogger: Subject<string>): Observable<string> {
    // init
    return of(aLayoutId).pipe(
        mergeMap(id => {
            // find the layout mapping
            const exMapping = _getLayoutMappingForType(aTypeId, aTools);
            // resolve
            const mappingFile: Observable<string> = exMapping ? empty() : _createLayoutMapping(aTools, aLogger);
            const mapping = exMapping ? of(exMapping) :
                mappingFile.pipe(mergeMap(file => rxReadFile(file)), map(data => JSON.parse(data)));
            // update the mapping file with the layout
            const newMapping = mapping.pipe(mergeMap(m => _addLayoutToMapping(id, m, aTools, aLogger)));

            // combine all
            return merge(mappingFile, newMapping).pipe(
                distinct());
        }));
}

function _writeContentItem(aName: string, aRoot: string, aDataDir: string, aState: State): Observable<string> {
    // id
    const id = v4();
    aState.contentItemId = id;
    // construct the target folder
    const targetFolder = rxMkdirp(join(aDataDir, 'content'));
    // load the content item
    const srcFile = join(__dirname, '..', '..', '..', 'assets', 'manifest.content.json');
    const fileName = `${id}_cmd.json`;
    // name
    const name = camelCase(`${aName} Web App Manifest`);
    // read
    return rxReadFile(srcFile).pipe(
        map(data => JSON.parse(data)),
        map(data => {
            // add the name
            data.id = id;
            data.name = name;
            // update the description
            data.elements.short_name.value = aName;
            data.elements.name.value = aName;
            // context root
            data.elements.context_root.value = aRoot;
            // inject the revision field
            return canonicalizeJSON(addRevisionId(data));
        }),
        mergeMap(data => targetFolder.pipe(
            mergeMap(tgt => rxWriteFile(join(tgt, fileName), JSON.stringify(data, undefined, 2))))));
}

const HEAD_REGEXP = /<\/head>/g;
const MANIFEST_REGEXP = /<link\s+[^>]*rel\s*=\s*\"manifest\"[^>]*>/g;
const FAVICON_REGEXP = /<link\s+[^>]*rel\s*=\s*\"icon\"[^>]*>/g;

function _fixIndex(aIndexContent: string): string {
    // the source
    const src = aIndexContent;
    // rewrite the favicon
    if (FAVICON_REGEXP.test(src)) {

    } else {
        // locate the end of the header and add the link

    }

    // rewrite the metadata
    /*    const manifest = `<link rel="manifest" href="${encodeURIComponent(MANIFEST_JSON)}">`;
        if (MANIFEST_REGEXP.test(aIndexContent)) {
            // replace
            src = src.replace(MANIFEST_REGEXP, manifest);
        } else {
            // locate the end of the header and add the links
            src = src.replace(HEAD_REGEXP, `${manifest}\n</head>`);
        } */

    // ok
    return src;
}

function _rewriteIndex(aSrcDir: string, aName: string, aLogger: Subject<string>): Observable<string> {
    // name
    const name = join(aSrcDir, 'index.html');
    return rxReadFile(name).pipe(
        map(idx => _fixIndex(idx)),
        mergeMap(data => rxWriteFile(name, data)));
}

// just copy the resources
function _copyResources(aDataDir: string): Observable<string> {
    // source
    const srcDir = join(__dirname, '..', '..', '..', 'assets', 'manifest');
    const dstDir = aDataDir;
    // copy
    return rxCopyDir(srcDir, dstDir);
}

/**
 * Copies the resources
 *
 * @param aName     name
 * @param aDataDir  data directory
 */
function _copyManifest(aName: string, aRoot: string, aDataDir: string, aState: State): Observable<string> {
    // copy the files
    const files = _copyResources(aDataDir);
    // rewritten
    const content = _writeContentItem(aName, aRoot, aDataDir, aState);
    // combine
    return merge(files, content).pipe(distinct());
}

function _rewritePackage(aSrcDir: string, aState: State): Observable<string> {
    // load the package json
    const filename = join(aSrcDir, 'package.json');
    // load
    return rxReadFile(filename).pipe(
        map(data => JSON.parse(data)),
        tap(data => data.config[KEY_MANIFEST_ID] = aState.contentItemId),
        map(data => JSON.stringify(data, undefined, 2)),
        mergeMap(data => rxWriteFile(filename, data))
    );
}

function _handleLayoutAndMapping(
    aRoot: string,
    aWchTools: WchTools,
    aLogger: Subject<string>): Observable<string> {
    // construct the mappings
    return concat(
        _writeLayoutAndMapping(aRoot, UUID_MANIFEST_TYPE, UUID_MANIFEST_LAYOUT, aWchTools, aLogger),
        _writeLayoutAndMapping(aRoot, UUID_MANIFEST_TYPE, UUID_FAVICONS_LAYOUT, aWchTools, aLogger)
    );
}

/**
 * Generates angular components based on type and layout mapping definitions
 *
 * create webappmanifest --data DATA_FOLDER
 */
export class CreateWebAppManifestCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'webappmanifest';

    static readonly COMMAND = CMD_CREATE;

    constructor(options: Options) {
        super(options);
    }

    process(): Observable<string> {
        // logger
        const logger = this.getLogger();
        // data directory
        const opts = this.getOptions();
        const dataDirOpt = this.getDataDir();
        const rxDataDir = this.rxMkdirp(dataDirOpt);
        // the directories
        const srcDirOpt = optNormalizeDir(firstOf(opts.src) || '.');
        const rxSrcDir = this.rxMkdirp(srcDirOpt);
        // root
        const root = opts.root || '';
        // relative or absolute
        const bRelative = !!opts.relative;
        // log the settings
        logger.next(`Source dir [${srcDirOpt}], relative [${bRelative}].`);
        // name
        const name = opts.name || 'Application';
        // state
        const state: State = {};
        // tools
        const rxWchTools = this.getWchTools();

        // manifest
        const rxManifest = concat(
            rxDataDir.pipe(mergeMap(aDataDir => _copyManifest(name, root, aDataDir, state))),
            rxSrcDir.pipe(mergeMap(aSourceDir => _rewritePackage(aSourceDir, state))));
        // layouts
        const rxLayouts = rxWchTools.pipe(mergeMap(aTools => _handleLayoutAndMapping(root, aTools, logger)));

        // combine
        return merge(rxManifest, rxLayouts);
    }
}
