"use strict";
exports.__esModule = true;
/* Copyright IBM Corp. 2018 */
var graceful_fs_1 = require("graceful-fs");
var mime_1 = require("mime");
var path_1 = require("path");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var verror_1 = require("verror");
var zlib_1 = require("zlib");
var rx_walk_1 = require("./rx.walk");
var brotli = require('brotli');
// register
mime_1.define({
    'application/brotli': ['br']
});
function _isCompressed(aMimeType) {
    return (aMimeType === 'application/gzip') || (aMimeType === 'application/gzip') || (aMimeType === 'application/brotli');
}
var GZIP_OPTIONS = {
    level: 9
};
function _readFile(aPath) {
    return rxjs_1.Observable.create(function (observer) {
        graceful_fs_1.readFile(aPath, function (errRead, data) {
            if (errRead) {
                observer.error(new verror_1.VError(errRead, 'Unable to read %s.', aPath));
            }
            else {
                observer.next(data);
                observer.complete();
            }
        });
    });
}
function _compressGzip(aFile, aData) {
    // compress
    return rxjs_1.Observable.create(function (aObserver) {
        // build the target file
        var parsed = path_1.parse(aFile);
        var dstFile = path_1.join(parsed.dir, parsed.base + '.gz');
        // compress
        zlib_1.gzip(aData, GZIP_OPTIONS, function (zipError, zipData) {
            if (zipError) {
                aObserver.error(new verror_1.VError(zipError, 'Unable to compress %s.', aFile));
            }
            else {
                graceful_fs_1.writeFile(dstFile, zipData, function (writeErr) {
                    if (writeErr) {
                        aObserver.error(new verror_1.VError(writeErr, 'Unable to write %s.', dstFile));
                    }
                    else {
                        aObserver.next(dstFile);
                        aObserver.complete();
                    }
                });
            }
        });
    });
}
var BROTLI_OPTIONS = {
    mode: 0,
    quality: 11,
    lgwin: 22 // window size
};
function _compressBrotli(aFile, aData) {
    // read the file
    var type = mime_1.getType(aFile);
    // otions
    var opts;
    // create the options
    if (type && type.startsWith('image/')) {
        opts = Object.assign({}, BROTLI_OPTIONS, { 'mode': 0 });
    }
    else {
        opts = Object.assign({}, BROTLI_OPTIONS, { 'mode': 1 });
    }
    // compress
    return rxjs_1.Observable.create(function (aObserver) {
        // build the target file
        var parsed = path_1.parse(aFile);
        var dstFile = path_1.join(parsed.dir, parsed.base + '.br');
        // compress
        var brotliData = brotli.compress(aData, opts);
        // write
        graceful_fs_1.writeFile(dstFile, brotliData, function (writeErr) {
            if (writeErr) {
                aObserver.error(new verror_1.VError(writeErr, 'Unable to write %s.', dstFile));
            }
            else {
                aObserver.next(dstFile);
                aObserver.complete();
            }
        });
    });
}
function _compressFile(aFile) {
    // test the mime type
    var type = mime_1.getType(aFile);
    if (type && _isCompressed(type)) {
        return rxjs_1.empty();
    }
    // read
    var rxRead = _readFile(aFile).pipe(operators_1.share());
    // gzip
    var rxGzip = rxRead.pipe(operators_1.mergeMap(function (data) { return _compressGzip(aFile, data); }));
    var rxBrotli = rxRead.pipe(operators_1.mergeMap(function (data) { return _compressBrotli(aFile, data); }));
    // compress
    return rxjs_1.merge(rxGzip, rxBrotli);
}
exports.rxCompressFile = _compressFile;
function _compressDir(aDir) {
    // walk the files
    return rx_walk_1.rxWalkFiles(aDir).pipe(operators_1.filter(function (fd) { return fd.stats.isFile(); }), operators_1.mergeMap(function (fd) { return _compressFile(fd.absPath); }));
}
exports.rxCompressDir = _compressDir;
