"use strict";
exports.__esModule = true;
var graceful_fs_1 = require("graceful-fs");
var rxjs_1 = require("rxjs");
var json_1 = require("./json");
var CRLF = /\r\n/g;
var CHARSET = 'utf-8';
/**
 * Replaces carriage return and line feeds with line feeds
 *
 * @param aValue    the string to replace
 * @return the normalized string
 */
function _normalizeNewline(aValue) {
    return aValue.replace(CRLF, '\n').trim();
}
function _readBinaryFile(aPath) {
    return rxjs_1.Observable.create(function (observer) {
        graceful_fs_1.readFile(aPath, function (errRead, data) {
            if (errRead) {
                observer.error(errRead);
            }
            else {
                observer.next(data);
                observer.complete();
            }
        });
    });
}
function _readTextFile(aPath) {
    return rxjs_1.Observable.create(function (observer) {
        graceful_fs_1.readFile(aPath, CHARSET, function (errRead, data) {
            if (errRead) {
                observer.error(errRead);
            }
            else {
                observer.next(data);
                observer.complete();
            }
        });
    });
}
exports.rxReadTextFile = _readTextFile;
function _writeTextFile(aPath, aData) {
    return rxjs_1.Observable.create(function (observer) { return _writeTextFileWithObserver(aPath, aData, observer); });
}
exports.rxWriteTextFile = _writeTextFile;
function _writeBinaryFile(aPath, aData) {
    return rxjs_1.Observable.create(function (observer) { return _writeBinaryFileWithObserver(aPath, aData, observer); });
}
var FILES = {};
function _writeTextFileWithObserver(aPath, aData, aObserver) {
    // validate that we do not write a file in parallel
    if (FILES[aPath]) {
        aObserver.error("Trying to write " + aPath + " in parallel.");
    }
    else {
        // register the file
        FILES[aPath] = aData;
        // execute the operation
        graceful_fs_1.writeFile(aPath, aData, CHARSET, function (errWrite) {
            // done
            delete FILES[aPath];
            // next
            if (errWrite) {
                aObserver.error(errWrite);
            }
            else {
                aObserver.next(aPath);
                aObserver.complete();
            }
        });
    }
}
function _stringDiffers(aLeft, aRight) {
    return aLeft !== aRight;
}
function _bufferDiffers(aLeft, aRight) {
    return !aLeft.equals(aRight);
}
function _writeBinaryFileWithObserver(aPath, aData, aObserver) {
    graceful_fs_1.writeFile(aPath, aData, function (errWrite) {
        if (errWrite) {
            aObserver.error(errWrite);
        }
        else {
            aObserver.next(aPath);
            aObserver.complete();
        }
    });
}
function _writeTextFileSafe(aPath, aData, bOverride) {
    // normalize the lines
    var normData = _normalizeNewline(aData);
    return rxjs_1.Observable.create(function (observer) {
        // first read the file
        graceful_fs_1.readFile(aPath, CHARSET, function (errRead, readData) {
            // if the file does not exist, just write
            if (errRead) {
                // write the file
                _writeTextFileWithObserver(aPath, normData, observer);
            }
            else {
                /**
                 * We have a file, check if we override
                 */
                if (bOverride) {
                    // only override if the data does not match
                    var normRead = _normalizeNewline(readData);
                    if (_stringDiffers(normData, normRead)) {
                        // content differs, write the file
                        _writeTextFileWithObserver(aPath, normData, observer);
                    }
                    else {
                        // nothing to produce, files are identical
                        observer.complete();
                    }
                }
                else {
                    // produce a gen file
                    var genPath_1 = aPath + '.$gen';
                    // check if the generated file differs
                    graceful_fs_1.readFile(genPath_1, CHARSET, function (errGen, genData) {
                        // gen file does not exist
                        if (errGen) {
                            // write the generated file
                            _writeTextFileWithObserver(genPath_1, normData, observer);
                        }
                        else {
                            // only override if the data does not match
                            var normRead = _normalizeNewline(genData);
                            if (_stringDiffers(normData, normRead)) {
                                // content differs, write the file
                                _writeTextFileWithObserver(genPath_1, normData, observer);
                            }
                            else {
                                // nothing to produce
                                observer.complete();
                            }
                        }
                    });
                }
            }
        });
    });
}
exports.writeTextFileSafe = _writeTextFileSafe;
function _writeBinaryFileSafe(aPath, aData, bOverride) {
    return rxjs_1.Observable.create(function (observer) {
        // first read the file
        graceful_fs_1.readFile(aPath, function (errRead, readData) {
            // if the file does not exist, just write
            if (errRead) {
                // write the file
                _writeBinaryFileWithObserver(aPath, aData, observer);
            }
            else {
                /**
                 * We have a file, check if we override
                 */
                if (bOverride) {
                    // only override if the data does not match
                    if (_bufferDiffers(aData, readData)) {
                        // content differs, write the file
                        _writeBinaryFileWithObserver(aPath, aData, observer);
                    }
                    else {
                        // nothing to produce, files are identical
                        observer.complete();
                    }
                }
                else {
                    // produce a gen file
                    var genPath_2 = aPath + '.$gen';
                    // check if the generated file differs
                    graceful_fs_1.readFile(genPath_2, function (errGen, genData) {
                        // gen file does not exist
                        if (errGen) {
                            // write the generated file
                            _writeBinaryFileWithObserver(genPath_2, aData, observer);
                        }
                        else {
                            // only override if the data does not match
                            if (_bufferDiffers(aData, genData)) {
                                // content differs, write the file
                                _writeBinaryFileWithObserver(genPath_2, aData, observer);
                            }
                            else {
                                // nothing to produce
                                observer.complete();
                            }
                        }
                    });
                }
            }
        });
    });
}
exports.writeBinaryFileSafe = _writeBinaryFileSafe;
function _writeJsonFileSafe(aPath, aData, bOverride) {
    // write the canonicalized file
    return _writeTextFileSafe(aPath, JSON.stringify(json_1.canonicalizeJSON(aData)), bOverride);
}
exports.writeJsonFileSafe = _writeJsonFileSafe;
