"use strict";
exports.__esModule = true;
var child_process_1 = require("child_process");
var crypto_1 = require("crypto");
var graceful_fs_1 = require("graceful-fs");
var lodash_1 = require("lodash");
var mime_1 = require("mime");
var os_1 = require("os");
var path_1 = require("path");
var process_1 = require("process");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var verror_1 = require("verror");
var config_1 = require("./../utils/config");
var assert_1 = require("./assert");
var json_1 = require("./json");
var rx_json_1 = require("./rx.json");
var rx_utils_1 = require("./rx.utils");
var rx_walk_1 = require("./rx.walk");
var LAYOUT_MAPPINGS = 'layout-mappings';
var LAYOUTS = 'layouts';
var TYPES = 'types';
var SITES = 'sites';
var CONTENT = 'content';
var IMAGE_PROFILES = 'image-profiles';
var ASSETS = 'assets';
var RENDITIONS = 'renditions';
function _sortPages(aLeft, aRight) {
    // check if we have data
    var posLeft = aLeft && aLeft.data && aLeft.data.position ? aLeft.data.position : 0, posRight = aRight && aRight.data && aRight.data.position ? aRight.data.position : 0;
    // compare
    return posLeft < posRight ? -1 : posLeft > posRight ? +1 : 0;
}
function _readSiteRecursive(aDir, aParent) {
    return rxjs_1.Observable.create(function (observer) {
        // check the type
        graceful_fs_1.readdir(aDir, function (err, files) {
            // map from name to structure
            var children = {};
            // handle the error
            if (err) {
                if (err.code === 'ENOENT') {
                    observer.complete();
                }
                else {
                    observer.error(err);
                }
            }
            else {
                // for each file test if it's a file or
                var allFiles = files.map(function (name) {
                    // get the status
                    var fileName = path_1.join(aDir, name);
                    return rx_walk_1.rxStats(fileName).pipe(operators_1.mergeMap(function (stats) {
                        if (stats.isDirectory()) {
                            // make sure we have a node of that name
                            var node = children[name] || (children[name] = {});
                            var ch = node.children || (node.children = []);
                            // recurse
                            return _readSiteRecursive(fileName, node);
                        }
                        else if (stats.isFile() &&
                            mime_1.getType(name) === 'application/json') {
                            // split the file part
                            var parsed = path_1.parse(name);
                            var pageName = parsed.name;
                            var node_1 = children[pageName] || (children[pageName] = {});
                            // load the json
                            return rx_walk_1.readFile(fileName).pipe(operators_1.map(function (data) { return JSON.parse(data); }), operators_1.map(function (data) { return (node_1.data = data); }), operators_1.map(function () { return node_1; }));
                        }
                        else {
                            return rxjs_1.empty();
                        }
                    }));
                });
                // wait for all
                rxjs_1.merge.apply(void 0, allFiles).pipe(operators_1.reduce(function (dst, src) {
                    return dst;
                }, children), operators_1.map(function (data) {
                    // validate the parent
                    var ch = aParent.children || (aParent.children = []);
                    // add all children
                    // tslint:disable-next-line:forin
                    for (var name_1 in children) {
                        ch.push(children[name_1]);
                    }
                    // order by position
                    ch.sort(_sortPages);
                }), operators_1.map(function () { return aParent; }))
                    .subscribe(observer);
            }
        });
    });
}
function _readSite(aDir, aSite) {
    // recurces
    return _readSiteRecursive(aDir, aSite);
}
function _isValidSite(aSite) {
    return aSite && aSite.data && aSite.data.id;
}
function _readSites(aDir, aMappings) {
    // construct a dummy helper
    var virtualRoot = {};
    return _readSite(aDir, virtualRoot).pipe(operators_1.map(function (root) {
        // reorder the sites into the mapping
        if (root.children) {
            root.children.filter(_isValidSite).forEach(function (site) {
                aMappings[site.data.id] = site;
            });
        }
    }), operators_1.toArray(), operators_1.map(function () { return aMappings; }));
}
function _readAll(aDir, aMappings, aNames) {
    // iterate over all files
    return rx_walk_1.walkFiles(aDir).pipe(operators_1.filter(rx_json_1.isJSON), operators_1.mergeMap(rx_json_1.readJSON), operators_1.reduce(function (dst, file) {
        dst[file.data.id] = file.data;
        aNames[file.data.id] = file.desc.absPath;
        return dst;
    }, aMappings));
}
function _read(aDir) {
    // read
    var files = {};
    var layoutMappings = _readAll(path_1.join(aDir, LAYOUT_MAPPINGS), {}, files);
    var layouts = _readAll(path_1.join(aDir, LAYOUTS), {}, files);
    var types = _readAll(path_1.join(aDir, TYPES), {}, files);
    var sites = _readSites(path_1.join(aDir, SITES), {});
    var content = _readAll(path_1.join(aDir, CONTENT), {}, files);
    var imageProfiles = _readAll(path_1.join(aDir, IMAGE_PROFILES), {}, files);
    // done
    return rxjs_1.combineLatest(layoutMappings, layouts, types, sites, content, imageProfiles).pipe(operators_1.map(function (_a) {
        var lm = _a[0], l = _a[1], t = _a[2], s = _a[3], c = _a[4], ip = _a[5];
        var tools = {
            root: aDir,
            layoutMappings: lm,
            layouts: l,
            types: t,
            idToFile: files,
            sites: s,
            content: c,
            imageProfiles: ip,
            rxAssets: rx_utils_1.rxCacheSingle(_readAll(path_1.join(aDir, ASSETS), {}, files)),
            rxRenditions: rx_utils_1.rxCacheSingle(_readAll(path_1.join(aDir, RENDITIONS), {}, files))
        };
        return tools;
    }));
}
exports.readWchTools = _read;
/**
 * Makes sure our path ends with a proper trailing slash
 */
function _ensureTrailingSlash(aUrl) {
    // test if the URL ends with a slash
    if (aUrl && aUrl.length > 0) {
        return aUrl[aUrl.length - 1] === '/' ? aUrl : aUrl + '/';
    }
    // nothing to do
    return aUrl;
}
exports.ensureTrailingSlash = _ensureTrailingSlash;
exports.KEY_API_URL = 'x-ibm-dx-tenant-base-url';
exports.KEY_USERNAME = 'username';
exports.FILE_WCHTOOLS = '.wchtoolsoptions';
exports.FILE_WCHTOOLS_JSON = '.wchtoolsoptions.json';
function _getApiBaseUrl(aOptions, aDataDir) {
    // verbose mode
    var bVerbose = !!aOptions.verbose;
    // check if we option
    var urlFromOptions = aOptions.url;
    if (urlFromOptions) {
        // some logging
        if (bVerbose) {
            console.log("Using URL [" + urlFromOptions + "] from options.");
        }
        return rxjs_1.of(_ensureTrailingSlash(urlFromOptions));
    }
    // check for the environment parameter
    var urlFromEnv = process_1.env[ENV_API_URL];
    if (urlFromEnv) {
        // some logging
        if (bVerbose) {
            console.log("Using URL [" + urlFromEnv + "] from environment.");
        }
        return rxjs_1.of(_ensureTrailingSlash(urlFromEnv));
    }
    // data
    var dataDir = aDataDir || aOptions.data;
    // test for the existencee of a local file
    var localFile = path_1.join(dataDir || '.', exports.FILE_WCHTOOLS), localJsonFile = path_1.join(dataDir || '.', exports.FILE_WCHTOOLS_JSON), homeFile = path_1.join(os_1.homedir(), exports.FILE_WCHTOOLS), homeJsonFile = path_1.join(os_1.homedir(), exports.FILE_WCHTOOLS_JSON);
    // try to load the tools file
    return rx_walk_1.readFile(localFile).pipe(operators_1.tap(function () {
        if (bVerbose) {
            console.log('read', localFile);
        }
    }), operators_1.catchError(function () {
        return rx_walk_1.readFile(localJsonFile).pipe(operators_1.tap(function () {
            if (bVerbose) {
                console.log('read', localJsonFile);
            }
        }));
    }), operators_1.catchError(function () {
        return rx_walk_1.readFile(homeFile).pipe(operators_1.tap(function () {
            if (bVerbose) {
                console.log('read', homeFile);
            }
        }));
    }), operators_1.catchError(function () {
        return rx_walk_1.readFile(homeJsonFile).pipe(operators_1.tap(function () {
            if (bVerbose) {
                console.log('read', homeJsonFile);
            }
        }));
    }), operators_1.map(function (data) { return JSON.parse(data); }), operators_1.pluck(exports.KEY_API_URL), operators_1.filter(assert_1.isValidUrl), operators_1.map(_ensureTrailingSlash));
}
exports.wchToolsGetApiBaseUrl = _getApiBaseUrl;
function _getRegExp(aString, aDefault) {
    if (aString) {
        // compile the pattern
        var pattern_1 = new RegExp(aString);
        return function (key) { return pattern_1.test(key); };
    }
    else {
        // always matches
        return function (key) { return aDefault; };
    }
}
function _getPattern(aOptions) {
    // get the pattern
    var include = _getRegExp(aOptions.include, true);
    var exclude = _getRegExp(aOptions.exclude, false);
    // return the combined function
    return function (key) { return include(key) && !exclude(key); };
}
exports.wchToolsGetMatcher = _getPattern;
function _loadPrivateKey() {
    // filename
    var name = path_1.join(os_1.homedir(), '.ssh', 'id_rsa');
    return rx_walk_1.readFile(name).pipe(operators_1.map(function (key) { return ({ key: key }); }));
}
var ENCRYPTED_ENCODING = 'base64';
var DECTYPTED_ENCODING = 'utf8';
function _encryptPassword(aPassword, aKey) {
    // encrypt
    return crypto_1.privateEncrypt(aKey, Buffer.from(aPassword, DECTYPTED_ENCODING)).toString(ENCRYPTED_ENCODING);
}
function _decryptPassword(aHash, aKey) {
    return crypto_1.privateDecrypt(aKey, Buffer.from(aHash, ENCRYPTED_ENCODING)).toString(DECTYPTED_ENCODING);
}
function _setCredentials(aWchToolsOptions, aKey, aData) {
    // validate the URL
    assert_1.assertHasTrailingSlash(aWchToolsOptions.baseUrl);
    // credentials
    var cred = {
        username: aWchToolsOptions.username,
        password: _encryptPassword(aWchToolsOptions.password, aKey)
    };
    // store
    aData[aWchToolsOptions.baseUrl] = cred;
    // ok
    return aData;
}
function _isValidCredential(aCred) {
    return !!(aCred &&
        assert_1.isValidUserName(aCred.username) &&
        lodash_1.isString(aCred.password));
}
function _emptyCredentials() {
    return {
        username: '',
        password: ''
    };
}
function _createCredentials(aUserName, aPassword) {
    return {
        username: aUserName,
        password: aPassword
    };
}
var ENV_USERNAME = 'ibm_wch_sdk_cli_username';
var ENV_PASSWORD = 'ibm_wch_sdk_cli_password';
var ENV_API_URL = 'ibm_wch_sdk_cli_url';
function _getCredentialsFromEnvironment() {
    // access the credentials from the environment
    var username = process_1.env[ENV_USERNAME] || '';
    var password = process_1.env[ENV_PASSWORD] || '';
    // construct
    return { username: username, password: password };
}
/**
 * Merge different credentials layers
 *
 * @param aBase         base layer
 * @param aOverride     override layer
 *
 * @return the merged credentials
 */
function _mergeCredentials(aBase, aOverride) {
    // target
    if (!aOverride) {
        return aBase;
    }
    // clone
    var cred = lodash_1.cloneDeep(aBase);
    // override
    if (!!aOverride.username) {
        cred.username = aOverride.username;
    }
    if (!!aOverride.password) {
        cred.password = aOverride.password;
    }
    // ok
    return cred;
}
function _loadCredentials(aApiBase) {
    // validate the URL
    assert_1.assertHasTrailingSlash(aApiBase);
    // credential file name
    var filename = path_1.join(os_1.homedir(), '.ibm-wch-sdk-cli', '.credentials');
    // read the credential
    var key = _loadPrivateKey();
    // load the file
    return rx_walk_1.readFile(filename).pipe(operators_1.map(function (data) { return JSON.parse(data); }), operators_1.map(function (data) { return data[aApiBase]; }), operators_1.mergeMap(function (cred) {
        return key.pipe(operators_1.map(function (k) { return _decryptPassword(cred.password, k); }), operators_1.map(function (p) {
            cred.password = p;
            return cred;
        }));
    }), operators_1.catchError(function () { return rxjs_1.of(_emptyCredentials()); }));
}
function _storeCredentials(aWchToolsOptions) {
    // target folder
    var folder = path_1.join(os_1.homedir(), '.ibm-wch-sdk-cli');
    var filename = path_1.join(folder, '.credentials');
    // read the credential
    var key = _loadPrivateKey();
    // create the directpry
    return rx_walk_1.rxMkdirp(folder).pipe(operators_1.mergeMap(function () { return rx_walk_1.readFile(filename).pipe(operators_1.map(function (data) { return JSON.parse(data); })); }), operators_1.catchError(function () { return rxjs_1.of({}); }), operators_1.mergeMap(function (data) {
        return key.pipe(operators_1.map(function (cred) { return _setCredentials(aWchToolsOptions, cred, data); }));
    }), operators_1.map(json_1.canonicalizeJSON), operators_1.map(function (data) { return JSON.stringify(data); }), operators_1.mergeMap(function (data) { return rx_walk_1.writeFile(filename, data); }));
}
exports.wchStoreCredentials = _storeCredentials;
function _getWindowsCredentials(aApiUrl) {
    // validate the URL
    assert_1.assertHasTrailingSlash(aApiUrl);
    // the executable
    var path = path_1.normalize(path_1.join(__dirname, '..', '..', '..', 'assets', 'credman', process.arch, 'WchCredMan.exe'));
    // execute
    var cmd = "\"" + path + "\" \"" + aApiUrl + "\"";
    // construct the observable
    return rxjs_1.Observable.create(function (observer) {
        // execute the command
        child_process_1.exec(cmd, {
            encoding: 'utf8'
        }, function (error, stdout, stderr) {
            if (error) {
                observer.error(error);
            }
            else {
                try {
                    // parse
                    observer.next(JSON.parse(stdout));
                    observer.complete();
                }
                catch (e) {
                    observer.error(e);
                }
            }
        });
    });
}
function _getStoredCredentials(aApiUrl) {
    // the key
    var key = _ensureTrailingSlash(aApiUrl);
    // normalize the URL
    if (os_1.platform() === 'win32') {
        // load the credentials module
        return _getWindowsCredentials(key).pipe(operators_1.mergeMap(function (cred) {
            return _isValidCredential(cred) ? rxjs_1.of(cred) : _loadCredentials(key);
        }), operators_1.catchError(function () { return _loadCredentials(key); }));
    }
    // linux like fallback
    return _loadCredentials(key);
}
function _getCredentials(aApiUrl) {
    // return
    return _getStoredCredentials(aApiUrl).pipe(operators_1.map(function (cred) { return _mergeCredentials(_getCredentialsFromEnvironment(), cred); }), operators_1.catchError(function (err) { return rxjs_1.of(_getCredentialsFromEnvironment()); }));
}
exports.wchToolsGetCredentials = _getCredentials;
function _getWchToolsOptions(aOptions, aDataDir) {
    // access the base URL
    return _getApiBaseUrl(aOptions, aDataDir).pipe(operators_1.mergeMap(function (url) {
        // the result
        var result;
        // get the credentials
        var user = aOptions.user;
        var password = aOptions.password;
        if (user && password) {
            // ok
            result = rxjs_1.of(_createCredentials(user, password));
        }
        else {
            // read credentials from the store
            result = _getCredentials(url);
        }
        // build the options
        return result.pipe(operators_1.map(function (cred) { return ({
            baseUrl: _ensureTrailingSlash(url),
            username: cred.username,
            password: cred.password
        }); }));
    }), 
    // make sure to share this
    operators_1.shareReplay(1));
}
exports.wchToolsGetOptions = _getWchToolsOptions;
function _hidePassword(aArgs) {
    // locate the password entry
    var copy = lodash_1.cloneDeep(aArgs);
    // find the pwd entry
    var idx = 0;
    while ((idx = copy.indexOf('--password', idx)) >= 0) {
        // replace next
        copy[++idx] = '***';
    }
    return copy;
}
exports.wchHidePassword = _hidePassword;
var noColors = /\x1B\[([0-9]{1,2}(;[0-9]{1,2})?)?[mGK]/g;
function _executeWchTools(aDataDir, aArgs, aOptions, aConfig, aLogger) {
    // sanity checks
    assert_1.assertNotNull(aDataDir, 'dataDir');
    assert_1.assertNotNull(aConfig, 'config');
    assert_1.assertIsValidUserName(aConfig.username, 'username');
    assert_1.assertNotNull(aConfig.password, 'password');
    assert_1.assertIsUrl(aConfig.baseUrl, 'apiUrl');
    // check if this is a delete command
    var bDelete = aArgs && aArgs.length > 0 && aArgs[0] === 'delete';
    // build the command line
    var args = [path_1.normalize(path_1.join(__dirname, 'fork.wchtools.js'))];
    args.push.apply(args, aArgs.concat(['--user',
        aConfig.username,
        '--password',
        aConfig.password,
        '--url',
        aConfig.baseUrl]));
    // add the directory
    if (!bDelete) {
        args.push('--dir', aDataDir);
    }
    // verbose flag
    if (aOptions.verbose) {
        args.push('--verbose');
    }
    // execute
    return rxjs_1.Observable.create(function (o) {
        // arguments
        var argString = JSON.stringify(_hidePassword(args));
        // log this
        aLogger.next("Spawning WCHTool with args [" + argString + "].");
        // spawn the process
        var childProcess = child_process_1.spawn(process.execPath, args, {
            cwd: aDataDir,
            env: Object.assign({
                NODE_TLS_REJECT_UNAUTHORIZED: '0'
            }, process_1.env)
        });
        // attach
        childProcess.stdout.on('data', function (data) {
            return aLogger.next(data.toString().replace(noColors, ''));
        });
        childProcess.stderr.on('data', function (data) {
            return aLogger.next(data.toString().replace(noColors, ''));
        });
        // end
        childProcess.on('close', function (code) {
            return code === 0
                ? o.complete()
                : o.error(new verror_1.VError('Exit code [%d] when executing wchtools with arguments [%s].', code, argString));
        });
        childProcess.on('error', function (err) {
            return o.error(new verror_1.VError(err, 'Unable to execute wchtools with arguments [%s].', argString));
        });
    });
}
exports.wchExecuteWchTools = _executeWchTools;
function _getDeliveryUrlFromApiUrl(aApiUrl) {
    // deduce the delivery URL from the api URL
    return aApiUrl.replace('/api/', '/');
}
exports.wchDeliveryUrlFromApiUrl = _getDeliveryUrlFromApiUrl;
function _createUrl(aContextRoot, aOpts) {
    // the delivery URL
    var absWchDeliveryUrl = _ensureTrailingSlash(_getDeliveryUrlFromApiUrl(aOpts.baseUrl));
    // path of the index file
    var absRootUrl = _ensureTrailingSlash("" + absWchDeliveryUrl + config_1.configTrimSlashes(aContextRoot));
    // append the index
    var url = absRootUrl + "index.html";
    // ok
    return rxjs_1.of(url);
}
exports.wchGetIndexUrl = _createUrl;
